package org.netkernel.http.session;

import java.util.*;
import java.util.concurrent.atomic.AtomicInteger;

import javax.servlet.http.Cookie;

import org.netkernel.container.IKernel;
import org.netkernel.http.rep.XSSSafeCookie;
import org.netkernel.http.transport.HTTPBridgeOverlay;
import org.netkernel.http.transport.HTTPRequestSpace2;
import org.netkernel.http.util.SecurityUtils;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFRequestReadOnly;
import org.netkernel.layer0.nkf.INKFResolutionContext;
import org.netkernel.layer0.nkf.INKFResponseReadOnly;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.layer0.util.Utils;
import org.netkernel.layer0.util.XMLReadable;
import org.netkernel.module.standard.StandardSpace;
import org.netkernel.module.standard.endpoint.TransparentOverlayImpl;
import org.netkernel.module.standard.endpoint.ConfiguredOverlayImpl.IConfig;
import org.netkernel.request.IRequestResponseFields;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.impl.RequestResponseFieldsImpl;
import org.netkernel.request.impl.RequestScopeLevelImpl;
import org.netkernel.urii.ISpace;
import org.w3c.dom.Document;

/*
 * Session Overlay
 */
public class HTTPSessionOverlay extends TransparentOverlayImpl
{
	private Map<String,Session> mSessions=new HashMap<String,Session>();

	private long mMinDuration=30000;
	private long mMaxDuration=300000;
	private int mMaxSessions=64;
	private String mCookiePath="/";
	private String mCookieName="NETKERNEL_SESSION_COOKIE";
	private boolean mAutoReconnect=true;
	private long mLastReap;	
	
	private AtomicInteger mSessionsCreated=new AtomicInteger();
	private AtomicInteger mSessionsExceeded=new AtomicInteger();
	
	private HTTPSessionSpace2Proto mProtoSpace;

	//private static final String NETKERNEL_SESSION_COOKIE=;
	//private HTTPSessionMetadataEndpoint mMetaEndpoint;
	
	public HTTPSessionOverlay()
	{	declareThreadSafe();
		declareUnhandledExceptionsExpired(false);
	}
	
	@Override
	public void onCommissionEndpoint(IKernel kernel, ISpace aSpace) throws Exception
	{	super.onCommissionEndpoint(kernel, aSpace);
		//mMetaEndpoint=new HTTPSessionMetadataEndpoint(this);
		//mMetaEndpoint.onCommissionEndpoint(kernel, context);
		
		mProtoSpace= new HTTPSessionSpace2Proto((StandardSpace)aSpace, this, kernel);
		mProtoSpace.onCommissionSpace(kernel);
	}

	@Override
	protected void postCommission(INKFRequestContext context)  throws Exception
	{	
		if (context.getParamValue("config")!=null)
		{
			Document config=(Document)context.source("param:config",Document.class);
			XMLReadable c=new XMLReadable(config);
			String s=c.getText("/config/maxDuration");
			if(!"".equals(s))
			{	mMaxDuration=Long.parseLong(s);				
			}
			s=c.getText("/config/minDuration");
			if(!"".equals(s))
			{	mMinDuration=Long.parseLong(s);				
			}
			s=c.getText("/config/maxSessions");
			if(!"".equals(s))
			{	mMaxSessions=Integer.parseInt(s);				
			}
			s=c.getText("/config/cookiePath");
			if(!"".equals(s))
			{	mCookiePath=s;				
			}
			s=c.getText("/config/cookieName");
			if(!"".equals(s))
			{	mCookieName=s;				
			}
			s=c.getText("/config/autoReconnect");
			if(!"".equals(s))
			{	mAutoReconnect=Boolean.parseBoolean(s);				
			}
		}
	}
	
	public ISpace[] getSpaces(INKFRequestContext aContext) throws Exception
	{	
		ISpace[] result = super.getSpaces(aContext);
		ISpace[] result2 = new ISpace[result.length+1];
		System.arraycopy(result, 0, result2, 1, result.length);
		result2[0]=mProtoSpace;
		return result2;
	}	
	
	public IHDSNode getSessionStatus()
	{	List<Session> sessions = new ArrayList(getSessions().values());
		HDSBuilder b=new HDSBuilder();
		b.pushNode("httpSessionStatus");
		b.addNode("sessions",sessions.size());
		int count=0;
		long now=System.currentTimeMillis();
		for (Session s: sessions)
		{	count+=s.getSessionContext(now).getItemCount();
		}
		b.addNode("totalObjects",count);
		b.addNode("meanObjectsPerSession",((float)count)/((float)sessions.size()));
		b.addNode("totalSessionsCreated",mSessionsCreated.get());
		b.addNode("totalSessionsExceeded",mSessionsExceeded.get());
		b.addNode("minDuration",getMinDuration());
		b.addNode("maxDuration",getMaxDuration());
		b.addNode("maxSessions",getMaxSessions());
		return b.getRoot();
	}
	
	public IRequestResponseFields getState()
	{	RequestResponseFieldsImpl fields=new RequestResponseFieldsImpl(super.getState());
		fields.put("status", getSessionStatus());
		return fields;
	}
	

	public void onRequest(String aID, INKFRequestContext context) throws Exception
	{	
		if (context.getThisRequest().getVerb()==INKFRequestReadOnly.VERB_META)
		{	Utils.delegateRequestInto(getDelegateSpace(),context);
		}
		else
		{	
			Long now=context.getKernelContext().getKernel().getTimer().getApproximateTime();
			
			//Swap cookie for session
			Cookie cookie=null;
			Session session=null;
			boolean create=true;
			boolean bad=false;
			HTTPSessionSpace sessionContext=null;
			cookie=(Cookie)context.source("httpRequest:/cookie/"+mCookieName, Cookie.class);
			if(cookie!=null)
			{	session=(Session)mSessions.get(cookie.getValue());
				if(session!=null)
				{	if ((now-session.getTouchTime())<mMaxDuration)
					{	sessionContext=session.getSessionContext(now);
						create=false;
					}
					else
					{	bad=true;
					}
				}
				else
				{	bad=true;
				}
			}
			
			if (bad)
			{	//cookie sent for non-existent or expired session, remove it throw exception
				if (mAutoReconnect)
				{	create=true;
				}
				else
				{	cookie=new Cookie(mCookieName, "");
					cookie.setPath(mCookiePath);
					cookie.setMaxAge(0);
					context.sink("httpResponse:/cookie",cookie);
					if (cookie!=null)
					{	mSessions.remove(cookie.getValue());
					}
					throw context.createFormattedException("EX_SESSION_EXPIRED", null, null, null);
				}
			}
			
			if(create)
			{	//reap first
				checkReap(context,now);				
				
				//No session so we need to create one
				String id;
				do
				{	id=SecurityUtils.GUID(null);
				}
				while(mSessions.get(id)!=null);
				sessionContext=new HTTPSessionSpace(getStandardSpace(),mProtoSpace,id);
				sessionContext.onCommissionSpace(context.getKernelContext().getKernel());				
				session=new Session(sessionContext,now);
				mSessionsCreated.incrementAndGet();
				mSessions.put(id,session);

				cookie=new XSSSafeCookie(mCookieName, id);
				//TODO Cookie should handle configurable path
				cookie.setPath(mCookiePath);
				context.sink("httpResponse:/cookie",cookie);
			}
			
			//Insert session context into superstack and target the delegate inner space.
			IRequestScopeLevel injectionContext = RequestScopeLevelImpl.createDurableScopeLevel(sessionContext,context.getKernelContext().getThisKernelRequest().getRequestScope());
			injectionContext=RequestScopeLevelImpl.createNonDurableScopeLevel(getDelegateSpace(), injectionContext);
			
			//Prepare sub-request into prepared context
			INKFRequest reqOut=context.getThisRequest().getIssuableClone();
			reqOut.setRequestScope(injectionContext);
			INKFResponseReadOnly resp=context.issueRequestForResponse(reqOut);
			
			//Send inner response as our response
			context.createResponseFrom(resp);
		}
	}
	
	protected Map getSessions()
	{	return mSessions;		
	}
	
	protected int getMaxSessions()
	{	return mMaxSessions;
	}
	
	protected long getMinDuration()
	{	return mMinDuration;
	}
	
	protected long getMaxDuration()
	{	return mMaxDuration;
	}
	
	private void checkReap(INKFRequestContext aContext, long aNow) throws Exception
	{	//reap at most once per 1024 milliseconds
		long next=aNow&(~0x3FF);
		if (mLastReap!=next)
		{	mLastReap=next;
			synchronized(this)
			{	reap(aContext,aNow,false);
				if (mSessions.size()>=mMaxSessions)
				{	reap(aContext,aNow,true);
				}
			}
		}
		if(mSessions.size()>=mMaxSessions)
		{	mSessionsExceeded.incrementAndGet();
			throw aContext.createFormattedException("EX_MAX_SESSIONS_EXCEEDED", "EX_MAX_SESSIONS_MESSAGE",null,null,mMaxSessions);
		}
	}
	
	private int reap(INKFRequestContext aContext, long aNow, boolean aForce) throws Exception
	{
		int cullSize=0;
		if (aForce)
		{	cullSize = mMaxSessions/4;
			if (cullSize==0) cullSize=1;
		}
		
		int cullCount=0;
		int size=mSessions.size();
		
		for (Iterator<Session> i=mSessions.values().iterator(); i.hasNext(); )
		{	Session session=i.next();
			long age = aNow-session.getTouchTime();
			if (age>mMaxDuration)
			{	cullCount++;
				i.remove();
			}
			else if (age>mMinDuration && cullCount<cullSize)
			{	cullCount++;
				i.remove();
			}
		}
		return cullCount;
	}
}